/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <1gaack@informatik.uni-hamburg.de> wrote this file. As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return Moritz Gaack
 * ----------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <math.h>
#include <glib.h>
#include <glib-object.h>

#include "ReferenceCountingImplJDG.h"
#include "GObjectMyObject.h"


/**
 * Typedefinition: TestFunction ist ein Funktionszeiger für zu testende Funktionen.
 *
 * @see performTest(TestFunction, unsigned long)
 */
typedef void (*TestFunction) (unsigned long);


/**
 * Typedefinition: MicroSeconds ist ein Datentyp zum Speichern von Mikrosekunden.
 *
 * @see getTimestamp()
 */
typedef unsigned long long MicroSeconds;


/**
 * Hilfsrountine, die ein einen Zeitstempel in Mikrosekunden ausgibt.
 *
 * Die Funktion berechnet aus dem timeval struct die vergangene Anzahl 
 * an Mikrosekunden und liefert diese zurück. Die Berechnung der Mikrosekunden
 * ist trivial und ergibt sich aus dem Aufbau des timeval structs.
 *
 * Siehe auch http://www.gnu.org/software/libc/manual/html_node/Elapsed-Time.html
 * Der struct timeval repräsentiert die vergangene Zeit. Deklaration in sys/time.h, sie 
 * enthält folgende Einträge:
 *    1. long int tv_sec
 *       Ganze Sekunden der vergangenen Zeit.
 *    2. long int tv_usec
 *       Der Rest der vergangenen Zeit in Mikrosekunden (Immer weniger als 1000000).
 *
 * @return Zeitstempel in Mikrosekunden
 */
static MicroSeconds getTimestamp()
{
    struct timeval now;
    gettimeofday (&now, NULL);
    return now.tv_usec + (MicroSeconds)now.tv_sec * 1000000;
}


/**
 * Hilfsrountine, die ein Performanceresultat ausgibt.
 *
 * Die Ausgabe besteht aus einer optionalen Nachricht, dem Timestamp zum Start und zum 
 * Ende des Tests. Die benötigte Anzahl an Millisekunden wird aus der Differenz aus End- 
 * und Startzeit berechnet und formatiert ausgegeben.
 *
 * @param message Optionale Nachricht, die im Logging mit ausgebgen wird.
 * @param start Timestamp zum Start des Tests.
 * @param end Timestamp zum Ende des Tests.
 */
static void printTimingResult(const char * message, MicroSeconds start, MicroSeconds end)
{
    double milliSeconds = (end - start) / 1000000.0L;
    if(message != NULL)
        printf("%s", message);
    printf("%4.6F s\n", milliSeconds);
}


/**
 * Testfunktion, die exemplarisch Allokation, Referenzierung und Deallokation untersucht.
 *
 * Diese Funktion verwendet keinen Overhead und dient daher als Maßstab für den Performancevergleich.
 * Die drei zu untersuchenden Funktionaliäten sind:
 *
 *   1. Allokation
 *      Es werden iterativ 'numIterations' int Variablen alloziert.
 *   2. Referenzierung
 *      Jede allozierte int Variable wird exemplarisch referenziert.
 *   3. Deallokation
 *      Reseverierter Speicher für 'numIterations' int Variablen wird iterativ freigegeben.
 *
 * @param numIterations Anzahl der Iterationen, die durchlaufen werden sollen
 * @see performTest(TestFunction, unsigned long)
 * @see testReferenceCountingImplJDG(unsigned long)
 * @see testReferenceCountingImplGLib(unsigned long)
 */
static void testWithoutReferenceCounting(unsigned long numIterations)
{
    MicroSeconds start, end;
    unsigned long i = 0;
    int * referencePointer = NULL;
    
    printf("TEST - Standard: Without Reference Counting..\n");
    //printf("Allocate buffer..\n");
    int **ptrs = (int**)calloc(numIterations, sizeof(int*));
    
    printf("Allocation (%lu): ", numIterations);
    start = getTimestamp();
    for(i = 0; i < numIterations; ++i)
    {
        ptrs[i] = (int *)calloc(1, sizeof(int));
    }
    end = getTimestamp();
    printTimingResult(NULL, start, end);
    
    printf("Retain (%lu): ", numIterations);
    start = getTimestamp();
    for(i = 0; i < numIterations; ++i)
    {
        referencePointer = ptrs[i];
    }
    end = getTimestamp();
    printTimingResult(NULL, start, end);
    
    printf("Release (%lu): ", numIterations);
    start = getTimestamp();
    for(i = 0; i < numIterations; ++i)
    {
        free(ptrs[i]);
    }
    end = getTimestamp();
    printTimingResult(NULL, start, end);
    
    //printf("Deallocate buffer..\n");
    free(ptrs);
}


/**
 * Testfunktion, die Allokation, Referenzierung und Deallokation nach einer Implementierung von
 * Jean-David Gadina untersucht.
 *
 * Diese Funktion verwendet einen minimalen Overhead und entspricht einer C-Implementierung der
 * Reference Counting Technik in Objective-C.
 *
 * Die drei zu untersuchenden Funktionaliäten sind:
 *
 *   1. Allokation
 *      Es werden iterativ 'numIterations' int Variablen (plus retainCount-Variable) alloziert.
 *   2. Referenzierung
 *      Jede allozierte int Variable wird exemplarisch referenziert, dabei wird der retainCount 
 *      inkrementiert.
 *   3. Deallokation
 *      Reseverierter Speicher für 'numIterations' int Variablen (plus zugehörige retainCount-Variable) 
 *      wird iterativ freigegeben. Der release-Aufruf ohne Speicherfreigabe wird nicht untersucht!
 *
 * @param numIterations Anzahl der Iterationen, die durchlaufen werden sollen
 * @see performTest(TestFunction, unsigned long)
 * @see testWithoutReferenceCounting(unsigned long)
 * @see testReferenceCountingImplGLib(unsigned long)
 */
static void testReferenceCountingImplJDG(unsigned long numIterations)
{
    MicroSeconds start, end;
    unsigned long i = 0;
    int * referencePointer = NULL;
    
    printf("TEST - Reference Counting Impl Jean-David Gadina..\n");
    //printf("Allocate buffer..\n");
    int ** ptrs = (int**)calloc(numIterations, sizeof(int *));
    
    printf("Allocation (%lu): ", numIterations);
    start = getTimestamp();
    for(i = 0; i < numIterations; ++i)
    {
        ptrs[i] = (int *)alloc(sizeof(int));
    }
    end = getTimestamp();
    printTimingResult(NULL, start, end);
    
    printf("Retain (%lu): ", numIterations);
    start = getTimestamp();
    for(i = 0; i < numIterations; ++i)
    {
        referencePointer = ptrs[i];
        retain(referencePointer);
    }
    end = getTimestamp();
    printTimingResult(NULL, start, end);
    
    // Necessary iteration, because otherwise performance comparison doesn't make sense.
    // In this iteration only the retaincount gets decreased - no allocated storage gets released.
    //printf("Decrease retain count..\n");
    for(i = 0; i < numIterations; ++i)
    {
        release(ptrs[i]);
    }
    
    printf("Release (%lu): ", numIterations);
    start = getTimestamp();
    for(i = 0; i < numIterations; ++i)
    {
        release(ptrs[i]);
    }
    end = getTimestamp();
    printTimingResult(NULL, start, end);
    
    //printf("Deallocate buffer..\n");
    free(ptrs);
}


/**
 * Testfunktion, die Allokation, Referenzierung und Deallokation nach einer Implementierung der
 * GLib-Bibliothek untersucht.
 *
 * Diese Funktion verwendet einen objektorientierten Ansatz in einer C-Implementierung. Die GLib-
 * Bibliothek unterstützt für alle GObject-Objekte Reference Counting.
 *
 * Die drei zu untersuchenden Funktionaliäten sind:
 *
 *   1. Allokation
 *      Es werden iterativ 'numIterations' int Variablen (plus benötigter Speicher für Objekt-
 *      kapselung) alloziert.
 *   2. Referenzierung
 *      Jede allozierte Klasse für eine int Variable wird exemplarisch referenziert, dabei wird 
 *      der retainCount inkrementiert.
 *   3. Deallokation
 *      Reseverierter Speicher für 'numIterations' int Variablen (plus benötigter Speicher für 
 *     Objektkapselung) wird iterativ freigegeben. Der release-Aufruf ohne Speicherfreigabe wird 
 *     nicht untersucht!
 *
 * @param numIterations Anzahl der Iterationen, die durchlaufen werden sollen
 * @see performTest(TestFunction, unsigned long)
 * @see testWithoutReferenceCounting(unsigned long)
 * @see testReferenceCountingImplJDG(unsigned long)
 */
static void testReferenceCountingImplGLib(unsigned long numIterations)
{
    MicroSeconds start, end;
    unsigned long i = 0;
    gpointer referencePointer = NULL;
    
    printf("TEST - Reference Counting Impl GLib..\n");
    //printf("Allocate buffer..\n");
    gpointer * ptrs = (gpointer*)calloc(numIterations, sizeof(gpointer));
    
    printf("Allocation (%lu): ", numIterations);
    start = getTimestamp();
    for(i = 0; i < numIterations; ++i)
    {
        ptrs[i] = my_object_new(0);
    }
    end = getTimestamp();
    printTimingResult(NULL, start, end);
    
    printf("Retain (%lu): ", numIterations);
    start = getTimestamp();
    for(i = 0; i < numIterations; ++i)
    {
        referencePointer = g_object_ref(G_OBJECT(ptrs[i]));
    }
    end = getTimestamp();
    printTimingResult(NULL, start, end);
    
    // Necessary iteration, because otherwise performance comparison doesn't make sense.
    // In this iteration only the retaincount gets decreased - no allocated storage gets released.
    //printf("Decrease retain count..\n");
    for(i = 0; i < numIterations; ++i)
    {
        g_object_unref(G_OBJECT(ptrs[i]));
    }
    
    printf("Release (%lu): ", numIterations);
    start = getTimestamp();
    for(i = 0; i < numIterations; ++i)
    {
        g_object_unref(G_OBJECT(ptrs[i]));
    }
    end = getTimestamp();
    printTimingResult(NULL, start, end);
    
    printf("Release buffer..\n");
    free(ptrs);
}


/**
 * Funktion, die einen Performancetest ausführt.
 *
 * Diese Funktion ruft die zu testende Funktion mit der 'numIterantions' Variable 
 * als Parameter auf. Eine Performancemessung ist für alle Funktionen möglich, 
 * die dem Typen 'TestFunction' entsprechen.
 *
 * Da diese Funktion eine weitere Indirektionsstufe für die zu messende Funktionalität 
 * bedeutet, wird die eigentliche Zeitmessung an die 'TestFunction' selbst delegiert.
 *
 * @param testFunction Zeiger auf die zu testende Funktion
 * @param numIterations Anzahl der Iterationen, die durchlaufen werden sollen
 * @see testWithoutReferenceCounting(unsigned long)
 * @see testReferenceCountingImplJDG(unsigned long)
 * @see testReferenceCountingImplGLib(unsigned long)
 */
static void performTest(TestFunction testFunction, unsigned long numIterations)
{
    /*
     MicroSeconds start, end;
     double milliSeconds = 0.0;
     */
    printf("==================================\n");
    
    /*
     start = getTimestamp();
     */
    testFunction(numIterations);
    /*
     end = getTimestamp();
     
     milliSeconds = (end - start) / 1000.0L;
     printf("time: %4.5f ms\n", milliSeconds);
     printf("==================================\n");
     */
}


/**
 * Funktion, die beim Programmstart aufgerufen wird.
 *
 * Diese Funktion entspricht der Hauptroutine des Programms. Sie gibt alle 
 * eingehenden Parameter aus, initialisiert die GLib-Bibliothek und startet 
 * alle Performance-Tests.
 *
 * @param argc Anzahl der Parameter
 * @param argv Ein Feld mit allen Parametern (Anzahl: #argc)
 */
int main(int argc, char * argv[])
{
    int i = 0;
    for (; i < argc; i++)
    {
        printf("argv[%d]: %s\n", i, argv[i]);
    }
    
    g_type_init();
    
    {
        unsigned long numIterations = 1000000;
        unsigned long x = 1;
        for(; x <= 10; ++x)
        {
            printf("Test with %lu iterations..\n", (x * numIterations));
            performTest(&testWithoutReferenceCounting, x * numIterations);
            performTest(&testReferenceCountingImplJDG, x * numIterations);
            performTest(&testReferenceCountingImplGLib, x * numIterations);
            printf("\n");
        }
    }

    return 0;
}

